@@ -1,4 +1,4 @@
-# Copyright (c) 2009-2012 Dennis Kaarsemaker <dennis@kaarsemaker.net>
+# Copyright (c) 2009-2014 Dennis Kaarsemaker <dennis@kaarsemaker.net>
 # A command-line interface to query the Django ORM
 #
 # Redistribution and use in source and binary forms, with or without modification,
@@ -27,8 +27,11 @@
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 from django.db.models import Q
-from django.core.management import BaseCommand
-from django.template import loader, Template, Context
+from django.db.models.base import ModelBase
+from django.db.models.fields import FieldDoesNotExist
+from django.core.exceptions import FieldError
+from django.core.management import BaseCommand, CommandError
+from django.template import loader, Template, Context, TemplateSyntaxError, TemplateDoesNotExist
 from django.conf import settings
 from optparse import make_option
 import os.path
@@ -37,7 +40,7 @@ import sys
 usage="""The django ORM will be queried with the filters on the commandline. Records
 will be separated with newlines, fields with the specified separator
 (the default is a comma). Alternatively, a template can be specified which
-will be passed the result of the query as the 'objects' variable
+will be passed the result of the query as the 'objects' variable.
 
 Query key/value pairs can be prefixed with a '!' or '~' to negate the query.
 The __in filter works, use a comma separated string of arguments
@@ -57,13 +60,8 @@ Examples:
  - Update the state of all mc2* servers
    $prog query -a servers -m Server name__startswith=mc2 -u status=live
 
-Operators you can filter with are listed on see
-https://docs.djangoproject.com/en/dev/ref/models/querysets/#field-lookups
-
-/!\\ Warning /!\\
-This script does not do much error checking. If you spell your query wrong, or
-do something wrong with templates, you will get a python traceback and not a
-nice error message."""
+Operators you can filter with are listed on
+https://docs.djangoproject.com/en/dev/ref/models/querysets/#field-lookups"""
 
 class Command(BaseCommand):
     option_list = BaseCommand.option_list + (
@@ -73,6 +71,8 @@ class Command(BaseCommand):
                     help="Query this model"),
         make_option('-f', '--fields', dest="fields", default=None,
                     help="Give these fields"),
+        make_option('-l', '--list-fields', dest="list_fields", default=False, action="store_true",
+                    help="List all available fields instead of running a query"),
         make_option('-o', '--order', dest="order", default=None,
                     help="Order by this field"),
         make_option('-s', '--separator', dest="separator", default=",",
@@ -89,76 +89,127 @@ class Command(BaseCommand):
 
     def handle(self, *args, **options):
         if not options['application']:
-            print "You must specify which application to use"
-            sys.exit(1)
+            raise CommandError("You must specify which application to use")
         if not options['model']:
-            print "You must specify which model to use"
-            sys.exit(1)
+            raise CommandError("You must specify which model to use")
         if not options['fields'] and not options['template'] and not options['template_file'] and not options['updates']:
-            print "You must specify a list of fields, a template or a set of updates"
-            sys.exit(1)
+            raise CommandError("You must specify a list of fields, a template or a set of updates")
 
         # Import the model
-        project_name = settings.SETTINGS_MODULE.split('.')[-2]
-        models = '.'.join([project_name, options['application'], 'models'])
-        __import__(models)
+        models = '%s.models' % options['application']
+        if options['application'] not in settings.INSTALLED_APPS:
+            raise CommandError("Application %s not found in INSTALLED_APPS. All applications: \n - %s" %
+                    (options['application'], "\n - ".join(settings.INSTALLED_APPS)))
+        try:
+            __import__(models)
+        except ImportError:
+            raise CommandError("Application %s could not be imported" % options['application'])
         models = sys.modules[models]
+        models_ = sorted([x for x in models.__dict__ if isinstance(models.__dict__[x], ModelBase)])
+        if options['model'] not in models_:
+            raise CommandError("Application %s has no %s model. Available models: \n - %s" %
+                    (options['application'], options['model'], "\n - ".join(models_)))
         model = getattr(models, options['model'])
 
+        # Help the user by displaying all fields if requested
+        if options['list_fields']:
+            self.stdout.write("Fields for %s.models.%s:\n - %s" % (options['application'], options['model'], '\n - '.join(model._meta.get_all_field_names())))
+            return
+
         # Create queryset from commandline arguments
         qargs = make_filter(args)
-        queryset = model.objects.filter(*qargs).distinct()
+        try:
+            queryset = model.objects.filter(*qargs).distinct()
+        except FieldError:
+            e = sys.exc_info()[1]
+            raise CommandError(str(e))
         if options['order']:
             queryset = queryset.order_by(options['order'])
 
         # Update
         if options['updates']:
+            for update in options['updates']:
+                if '=' not in update:
+                    raise CommandError("Invalid update statement: %s" % update)
             updates = dict([x.split('=', 1) for x in options['updates']])
+
             for key in updates:
-                choices = [x[0] for x in model._meta.get_field_by_name(key)[0].choices]
+                valid_fields = [x[0].name for x in model._meta.get_fields_with_model()]
+                try:
+                    field = model._meta.get_field_by_name(key)[0]
+                    if key not in valid_fields:
+                        raise FieldDoesNotExist("Field %s is not a direct field." % key)
+                    choices = [x[0] for x in model._meta.get_field_by_name(key)[0].choices]
+                except FieldDoesNotExist:
+                    e = "%s. Choices are: %s" % (str(sys.exc_info()[1]), ', '.join(valid_fields))
+                    raise CommandError(e)
                 if choices and updates[key] not in choices:
                     raise ValueError("Invalid choice for %s: %s. Valid choices: %s" % (key, updates[key], ', '.join(choices)))
             keylen = max([len(x) for x in updates]) + 3
             vallen = max([len(str(getattr(obj, key))) for obj in queryset for key in updates]) + 3
             for obj in queryset:
-                print str(obj)
+                self.stdout.write(str(obj))
                 for key in sorted(updates.keys()):
                     sys.stdout.write('  ' +  key + ' ' * (keylen - len(key)))
                     sys.stdout.write(str(getattr(obj, key)) + ' ' * (vallen - len(str(getattr(obj,key)))))
                     sys.stdout.write('=> ' + updates[key] + "\n")
             resp = raw_input("Apply changes? [y/N] ")
             if resp.lower() != 'y':
-                print "Aborted"
+                self.stdout.write("Aborted")
             else:
                 queryset.update(**updates)
-                print "Applied!"
+                self.stdout.write("Applied!")
 
         # Generate output
-        if options['template'] or options['template_file']:
-            template = Template(options['template'])
+        if options['template']:
+            try:
+                template = Template(options['template'])
+            except TemplateSyntaxError:
+                raise CommandError("Syntax error in template: %s" % str(sys.exc_info()[1]))
+            self.stdout.write(template.render(Context({'objects': queryset})))
+        if options['template_file']:
             tf = options['template_file']
             if tf == '-':
                 template = Template(sys.stdin.read())
             elif tf and os.path.exists(tf):
-                template = Template(open(tf).read())
+                try:
+                    template = Template(open(tf).read())
+                except TemplateSyntaxError:
+                    raise CommandError("Syntax error in template: %s" % str(sys.exc_info()[1]))
             elif tf:
-                template = loader.get_template(tf)
-            print template.render(Context({'objects': queryset}))
+                try:
+                    template = loader.get_template(tf)
+                except TemplateDoesNotExist:
+                    raise CommandError("Cannot find a template named %s" % tf)
+                except TemplateSyntaxError:
+                    raise CommandError("Syntax error in template: %s" % str(sys.exc_info()[1]))
+            self.stdout.write(template.render(Context({'objects': queryset})))
         elif options['fields']:
             def getattr_r(obj, attr):
                 if '.' in attr:
                     me, next = attr.split('.',1)
-                    return getattr_r(getattr(obj, me), next)
-                return getattr(obj, attr)
+                    try:
+                        return getattr_r(getattr(obj, me), next)
+                    except AttributeError:
+                        e = "%s. Choices are: %s" % (str(sys.exc_info()[1]), ', '.join(sorted([x for x in dir(obj) if not x.startswith('_')])))
+                        raise CommandError(e)
+                try:
+                    return getattr(obj, attr)
+                except AttributeError:
+                    e = "%s. Choices are: %s" % (str(sys.exc_info()[1]), ', '.join(sorted([x for x in dir(obj) if not x.startswith('_')])))
+                    raise CommandError(str(e))
+
             fields = options['fields'].split(',')
             for record in queryset:
-                print options['separator'].join([unicode(getattr_r(record, x)) for x in fields])
+                self.stdout.write(options['separator'].join([unicode(getattr_r(record, x)) for x in fields]))
 
 def make_filter(args):
     qargs = []
     for x in args:
         if not x:
             continue
+        if '=' not in x:
+            raise CommandError("Invalid filter: %s" % x)
         key, val = x.split('=',1)
         if key.endswith('__in'):
             val = val.split(',')
